/**
* Copyright (C) 2007 EDIT
* European Distributed Institute of Taxonomy 
* http://www.e-taxonomy.eu
* 
* The contents of this file are subject to the Mozilla Public License Version 1.1
* See LICENSE.TXT at the top of this package for the full license terms.
*/

package eu.etaxonomy.taxeditor.editor;

import java.net.URL;
import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Logger;
import org.eclipse.core.runtime.Assert;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.ui.forms.IManagedForm;
import org.eclipse.ui.forms.ManagedForm;

import eu.etaxonomy.cdm.model.description.CommonTaxonName;
import eu.etaxonomy.cdm.model.description.DescriptionElementBase;
import eu.etaxonomy.cdm.model.description.Feature;
import eu.etaxonomy.cdm.model.description.TaxonDescription;
import eu.etaxonomy.cdm.model.name.HomotypicalGroup;
import eu.etaxonomy.cdm.model.taxon.Synonym;
import eu.etaxonomy.cdm.model.taxon.Taxon;
import eu.etaxonomy.cdm.model.taxon.TaxonRelationship;
import eu.etaxonomy.taxeditor.controller.EditorController;
import eu.etaxonomy.taxeditor.editor.description.CommonNameComposite;
import eu.etaxonomy.taxeditor.editor.description.DescriptionElementComposite;
import eu.etaxonomy.taxeditor.editor.description.DescriptionFeatureGroupComposite;
import eu.etaxonomy.taxeditor.editor.description.DescriptionGroupComposite;
import eu.etaxonomy.taxeditor.editor.description.DescriptionLabelComposite;
import eu.etaxonomy.taxeditor.editor.images.ImageComposite;
import eu.etaxonomy.taxeditor.editor.images.ImagesGroupComposite;
import eu.etaxonomy.taxeditor.editor.images.TaxonHasNoImagesComposite;
import eu.etaxonomy.taxeditor.editor.name.AcceptedNameComposite;
import eu.etaxonomy.taxeditor.editor.name.ConceptComposite;
import eu.etaxonomy.taxeditor.editor.name.ConceptGroupComposite;
import eu.etaxonomy.taxeditor.editor.name.HomotypicalGroupComposite;
import eu.etaxonomy.taxeditor.editor.name.MisappliedGroupComposite;
import eu.etaxonomy.taxeditor.editor.name.MisappliedNameComposite;
import eu.etaxonomy.taxeditor.editor.name.SynonymComposite;

/**
 * @author p.ciardelli
 * @created 08.01.2009
 * @version 1.0
 */
public class FreeTextElementFactory implements IFreeTextElementFactory {
	private static final Logger logger = Logger
			.getLogger(FreeTextElementFactory.class);
	
	private static IFreeTextElementFactory factory;
	
	private Map<Taxon, IManagedForm> nameForms;
	private Map<Taxon, IManagedForm> descriptionForms;

	private HashMap<Taxon, IManagedForm> imagesForms;
	
	public static IFreeTextElementFactory getDefault() {
		if (factory == null) {
			factory = new FreeTextElementFactory();
		}
		return factory;
	}
	
	public Composite createAcceptedTaxon(IManagedForm form, Taxon taxon) {
		
		// Put the taxon's form in the form map
		Assert.isNotNull(form, "A non-null IManagedForm must be provided.");
		getTaxonNameForms().put(taxon, form);
		
		// Create a homotypic group composite for the accepted taxon
		HomotypicalGroup group = taxon.getHomotypicGroup();
		Assert.isNotNull(group, "Taxon does not have a homotypic group");
		Composite groupComposite = createHomotypicalGroup(taxon, group);
				
		// Create a name composite for the accepted taxon
		Composite acceptedNameComposite =
				new AcceptedNameComposite(groupComposite, form, taxon);
	
		// Return the name composite
		return acceptedNameComposite;
	}
	
	public Composite createSynonym(Taxon taxon, Synonym synonym) {
		
		// Get the taxon's form
		IManagedForm form = getNameForm(taxon);
		Assert.isNotNull(form, "No IManagedForm has been provided for the synonym's accepted taxon.");
		
		// Get the synonym's homotypic group
		HomotypicalGroup group = synonym.getHomotypicGroup();
		Assert.isNotNull(group, "Synonym does not have a homotypic group");
		
		// If the group doesn't yet have a composite, create one and add it to the repository
		Composite groupComposite = 
				EditorController.getNameEditor(taxon).getHomotypicGroup(group);
		if (groupComposite == null) {
			groupComposite = createHomotypicalGroup(taxon, group);
		}
				
		// Create a synonym composite in the homotypic group
		Composite synonymComposite = 
				new SynonymComposite(groupComposite, form, taxon, synonym);
				
		return synonymComposite;
	}
	
	
	public Composite createMisappliedName(Taxon taxon, Taxon misappliedName) {

		// Get the taxon's form
		IManagedForm form = getNameForm(taxon);
		Assert.isNotNull(form, "No IManagedForm has been provided for the synonym's accepted taxon.");
				
		// If there is no composite for misapplied names, 
		//	create one and add it to the repository
		Composite groupComposite = 
				EditorController.getNameEditor(taxon).getMisappliedGroup();
		if (groupComposite == null) {
			groupComposite = createMisappliedGroup(taxon);
		}
		
		// Create the name's composite
		Composite composite = new MisappliedNameComposite
				(groupComposite, form, taxon, misappliedName);
		
		return composite;
	}
	
	
	public Composite createMisappliedGroup(Taxon taxon) {
		
		// Get the taxon's form
		IManagedForm form = getNameForm(taxon);
		Assert.isNotNull(form, "No IManagedForm has been provided for the misapplied group's accepted taxon.");

		// Create the group composite
		Composite groupComposite = new MisappliedGroupComposite(getTopLevelComposite(form), taxon);
		groupComposite.addFocusListener(new CompositeBorderDecorator(groupComposite, form));
		
		// Put the group composite before concept group composite, if any
		Composite conceptGroupComposite = 
			EditorController.getNameEditor(taxon).getConceptGroup();
		if (conceptGroupComposite != null) {
			groupComposite.moveAbove(conceptGroupComposite);
		}
		
		return groupComposite;
	}
	
	public Composite createHomotypicalGroup(Taxon taxon, HomotypicalGroup group) {

		// Get the taxon's form
		IManagedForm form = getNameForm(taxon);
		Assert.isNotNull(form, "No IManagedForm has been provided for the homotypical group's accepted taxon.");

		// Create the group composite
		Composite groupComposite = new HomotypicalGroupComposite(getTopLevelComposite(form), taxon, group);
		groupComposite.addFocusListener(new CompositeBorderDecorator(groupComposite, form));		
				
		return groupComposite;
	}

	private Map<Taxon, IManagedForm> getTaxonNameForms() {
		if (nameForms == null) {
			nameForms = new HashMap<Taxon, IManagedForm>();
		}
		return nameForms;
	}
	
	private IManagedForm getNameForm(Taxon taxon) {
		return getTaxonNameForms().get(taxon);
	}
		
	private Composite getTopLevelComposite(IManagedForm form) {
		return form.getForm().getBody();
	}

	
	public Composite createConcept(Taxon taxon, TaxonRelationship relationship) {
		// Get the taxon's form
		IManagedForm form = getNameForm(taxon);
		Assert.isNotNull(form, "No IManagedForm has been provided for the synonym's accepted taxon.");
				
		// If there is no composite for misapplied names, 
		//	create one and add it to the repository
		Composite groupComposite = 
				EditorController.getNameEditor(taxon).getConceptGroup();
		if (groupComposite == null) {
			groupComposite = createConceptGroup(taxon);
		}
		
		// Create the name's composite
		Composite composite = ConceptComposite.getNewInstance(groupComposite, form, 
									taxon, relationship);
				
		return composite;
	}

	
	public Composite createConceptGroup(Taxon taxon) {
		
		// Get the taxon's form
		IManagedForm form = getNameForm(taxon);
		Assert.isNotNull(form, "No IManagedForm has been provided for the concept group's accepted taxon.");

		// Create the group composite
		Composite groupComposite = new ConceptGroupComposite(getTopLevelComposite(form), taxon);
		groupComposite.addFocusListener(new CompositeBorderDecorator(groupComposite, form));
		
		// Put the group composite after misapplied group composite, if any
		Composite misappliedGroupComposite = 
			EditorController.getNameEditor(taxon).getMisappliedGroup();
		if (misappliedGroupComposite != null) {
			groupComposite.moveBelow(misappliedGroupComposite);
		}
		
		return groupComposite;

	}

	
	public Composite createDescriptionLabel(IManagedForm form, Taxon taxon,
			TaxonDescription description) {
		
		// Put the taxon's form in the form map if no other description has done so
		Assert.isNotNull(form, "A non-null IManagedForm must be provided.");
		getTaxonDescriptionForms().put(taxon, form);
		
		// Create a group composite for the description
		Composite groupComposite = createDescriptionGroup(taxon, description);
				
		// Create label composite
		Composite labelComposite =
				new DescriptionLabelComposite(groupComposite, form, description);
	
		// Return the label composite
		return labelComposite;
	}

	
	public Composite createDescriptionElement(Taxon taxon, TaxonDescription description,
			DescriptionElementBase element) {
		
		// Get the taxon's form
		IManagedForm form = getDescriptionForm(taxon);
		Assert.isNotNull(form, "No IManagedForm has been provided for the element's taxon.");

		// Get the element's feature
		Feature feature = element.getFeature();
		Assert.isNotNull(feature, "Element does not have a feature.");
							
		// If the feature doesn't yet have a composite, create one and add it to the repository
		Composite groupComposite = EditorController.getDescriptionEditor(taxon)
										.getDescriptionFeatureGroup(description, feature);
		if (groupComposite == null) {
			groupComposite = createDescriptionFeatureGroup(taxon, description, feature);
		}
				
		// Create an element composite in the feature group according to its type
		Composite elementComposite = null;
		if (element.isInstanceOf(CommonTaxonName.class)) {
			elementComposite =
					new CommonNameComposite(groupComposite, form, (CommonTaxonName) element);
		} else {
			elementComposite = 
					new DescriptionElementComposite(groupComposite, form, element);
		}
				
		return elementComposite;
	}

	
	public Composite createDescriptionGroup(Taxon taxon,
			TaxonDescription description) {
		
		// Get the taxon's form
		IManagedForm form = getDescriptionForm(taxon);
		Assert.isNotNull(form, "No IManagedForm has been provided for the description's taxon.");

		// Create the group composite
		GroupComposite groupComposite = new DescriptionGroupComposite(getTopLevelComposite(form), taxon, description);
		groupComposite.addFocusListener(new CompositeBorderDecorator(groupComposite, form));
		groupComposite.setDroppable(false);
				
		return groupComposite;
	}
		
	
	public Composite createDescriptionFeatureGroup(Taxon taxon,
			TaxonDescription description, Feature feature) {

		// Get the taxon's form
		IManagedForm form = getDescriptionForm(taxon);
		Assert.isNotNull(form, "No IManagedForm has been provided for the feature's taxon.");
		
		// Description label should be created before adding features
		Composite descriptionGroupComposite = 
			EditorController.getDescriptionEditor(taxon).getDescriptionGroup(description);
		Assert.isNotNull(descriptionGroupComposite, 
				"Description group must be created - usually by creating its label - before adding features.");
		
		// Create the group composite
		Composite groupComposite = new DescriptionFeatureGroupComposite
										(descriptionGroupComposite, taxon, description, feature);
		groupComposite.addFocusListener(new CompositeBorderDecorator(groupComposite, form));	
		
		return groupComposite;		
	}
	
	private Map<Taxon, IManagedForm> getTaxonDescriptionForms() {
		if (descriptionForms == null) {
			descriptionForms = new HashMap<Taxon, IManagedForm>();
		}
		return descriptionForms;
	}
	

	private IManagedForm getDescriptionForm(Taxon taxon) {
		return getTaxonDescriptionForms().get(taxon);
	}

	/* (non-Javadoc)
	 * @see eu.etaxonomy.taxeditor.editor.IFreeTextElementFactory#createImageElement(org.eclipse.ui.forms.ManagedForm, eu.etaxonomy.cdm.model.taxon.Taxon, java.net.URL)
	 */
	public Composite createImageElement(ManagedForm form, Taxon taxon, URL url) {
				
		// Get the taxon's form
		Assert.isNotNull(form, "A non-null IManagedForm must be provided.");

		// Get the group composite for all images
		GroupComposite groupComposite = 
				EditorController.getImageEditor(taxon).getImageGroup();
		if (groupComposite == null) {
			groupComposite = createImagesGroup(taxon, form);
		}
		
		// Create the image composite
		GroupedComposite imageComposite = new ImageComposite(groupComposite, form, taxon, url);
		imageComposite.addFocusListener(new CompositeBorderDecorator(imageComposite, form));
				
		return imageComposite;
	}

	/**
	 * @param taxon
	 * @param form2 
	 * @return
	 */
	private GroupComposite createImagesGroup(Taxon taxon, ManagedForm form) {

		// Create the group composite
		GroupComposite groupComposite = new ImagesGroupComposite(getTopLevelComposite(form), taxon);		
				
		return groupComposite;
	}

	/**
	 * @param taxon
	 * @return
	 */
	private IManagedForm getImagesForm(Taxon taxon) {
		return getTaxonImagesForms().get(taxon);
	}

	/**
	 * @return
	 */
	private Map<Taxon, IManagedForm> getTaxonImagesForms() {
		if (imagesForms == null) {
			imagesForms = new HashMap<Taxon, IManagedForm>();
		}
		return imagesForms;
	}

	/* (non-Javadoc)
	 * @see eu.etaxonomy.taxeditor.editor.IFreeTextElementFactory#createTaxonHasNoImagesElement(org.eclipse.ui.forms.ManagedForm, eu.etaxonomy.cdm.model.taxon.Taxon)
	 */
	public Composite createTaxonHasNoImagesElement(ManagedForm form,
			Taxon taxon) {
		// Get the taxon's form
		Assert.isNotNull(form, "A non-null IManagedForm must be provided.");

		// Get the group composite for all images
		GroupComposite groupComposite = 
				EditorController.getImageEditor(taxon).getImageGroup();
		if (groupComposite == null) {
			groupComposite = createImagesGroup(taxon, form);
		}
		
		// Create the image composite
		GroupedComposite imageComposite = new TaxonHasNoImagesComposite(groupComposite, form, taxon);
				
		return imageComposite;		
	}
}
