/**
 * Copyright (C) 2007 EDIT
 * European Distributed Institute of Taxonomy 
 * http://www.e-taxonomy.eu
 * 
 * The contents of this file are subject to the Mozilla Public License Version 1.1
 * See LICENSE.TXT at the top of this package for the full license terms.
 */

package eu.etaxonomy.taxeditor.model;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.log4j.Logger;
import org.eclipse.jface.dialogs.MessageDialog;

import eu.etaxonomy.cdm.api.application.CdmApplicationController;
import eu.etaxonomy.cdm.api.service.INameService;
import eu.etaxonomy.cdm.api.service.IReferenceService;
import eu.etaxonomy.cdm.api.service.ITaxonService;
import eu.etaxonomy.cdm.api.service.config.ITaxonServiceConfigurator;
import eu.etaxonomy.cdm.api.service.config.impl.TaxonServiceConfiguratorImpl;
import eu.etaxonomy.cdm.common.CdmUtils;
import eu.etaxonomy.cdm.model.common.IdentifiableEntity;
import eu.etaxonomy.cdm.model.common.TimePeriod;
import eu.etaxonomy.cdm.model.name.HomotypicalGroup;
import eu.etaxonomy.cdm.model.name.NameRelationship;
import eu.etaxonomy.cdm.model.name.NameRelationshipType;
import eu.etaxonomy.cdm.model.name.Rank;
import eu.etaxonomy.cdm.model.name.TaxonNameBase;
import eu.etaxonomy.cdm.model.reference.ReferenceBase;
import eu.etaxonomy.cdm.model.taxon.Synonym;
import eu.etaxonomy.cdm.model.taxon.SynonymRelationshipType;
import eu.etaxonomy.cdm.model.taxon.Taxon;
import eu.etaxonomy.cdm.model.taxon.TaxonBase;
import eu.etaxonomy.cdm.persistence.query.MatchMode;
import eu.etaxonomy.taxeditor.TaxEditorPlugin;
import eu.etaxonomy.taxeditor.controller.GlobalController;
import eu.etaxonomy.taxeditor.navigation.SearchResult;

/**
 * @author p.ciardelli
 * @created 17.09.2008
 * @version 1.0
 */
public class CdmUtil {
	private static final Logger logger = Logger.getLogger(CdmUtil.class);
	
	/**
	 * Checks whether synonym's name is the basionym for ALL names 
	 * 	in its group.
	 * 
	 * @param synonym
	 * 
	 * @return
	 */
	public static boolean isSynonymGroupBasionym(Synonym synonym) {
		
		TaxonNameBase synonymName = synonym.getName();
		return isNameGroupBasionym(synonymName);
	}
	
	/**
	 * Checks whether name is the basionym for ALL names 
	 * 	in its group.
	 * @param name
	 * 
	 * @return
	 */	
	public static boolean isNameGroupBasionym(TaxonNameBase name) {
		if (name == null) {
			return false;
		}
		
		HomotypicalGroup homotypicalGroup = name.getHomotypicalGroup();
		if (homotypicalGroup == null) {
			return false;
		}		
		
		Set<TaxonNameBase> typifiedNames = homotypicalGroup.getTypifiedNames();
		
		// Check whether there are any other names in the group
		if (typifiedNames.size() == 1) {
			return false;
		}
		
		boolean isBasionymToAll = true;
				
		for (TaxonNameBase taxonName : typifiedNames) {
			if (!taxonName.equals(name)) {
				if (!isNameBasionymOf(name, taxonName)) {
					return false;
				}
			}
		}
		return true;		
	}
	
	/**
	 * Checks whether a basionym relationship exists between fromName and toName.
	 * 
	 * @param fromName
	 * @param toName
	 * @return
	 */
	public static boolean isNameBasionymOf(TaxonNameBase fromName, TaxonNameBase toName) {
		Set<NameRelationship> relations = toName.getRelationsToThisName();
		for (NameRelationship relation : relations) {
			if (relation.getType().equals(NameRelationshipType.BASIONYM()) &&
					relation.getFromName().equals(fromName)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Creates a basionym relationship between basionymName and
	 * 	each name in its homotypic group.
	 * 
	 * @param basionymName
	 */
	public static void setGroupBasionym(TaxonNameBase basionymName) {
		HomotypicalGroup homotypicalGroup = basionymName.getHomotypicalGroup();
		if (homotypicalGroup == null) {
			return;
		}
		for (TaxonNameBase name : homotypicalGroup.getTypifiedNames()) {
			if (!name.equals(basionymName)) {
				
				// First check whether the relationship already exists
				if (!isNameBasionymOf(basionymName, name)) {
					
					// Then create it
					name.addRelationshipFromName(basionymName, 
							NameRelationshipType.BASIONYM(), null);
				}
			}
		}
	}

	/**
	 * Removes all basionym relationships between basionymName and
	 * 	the names in its homotypic group.
	 * 
	 * @param basionymName
	 */
	public static void removeGroupBasionym(TaxonNameBase basionymName) {
		HomotypicalGroup homotypicalGroup = basionymName.getHomotypicalGroup();
		Set<NameRelationship> relations = basionymName.getRelationsFromThisName();
		Set<NameRelationship> removeRelations = new HashSet<NameRelationship>();
		
		for (NameRelationship relation : relations) {
			
			// If this is a basionym relation, and toName is in the homotypical group,
			//	remove the relationship.
			if (relation.getType().equals(NameRelationshipType.BASIONYM()) &&
					relation.getToName().getHomotypicalGroup().equals(homotypicalGroup)) {
				removeRelations.add(relation);
			}
		}
		
		// Removing relations from a set through which we are iterating causes a 
		//	ConcurrentModificationException. Therefore, we delete the targeted
		//	relations in a second step.
		for (NameRelationship relation : removeRelations) {
			basionymName.removeNameRelationship(relation);
		}
	}
	
	/**
	 * Checks whether name belongs to the same homotypic group as taxon's name.
	 * 
	 * @param name
	 * @param taxon
	 * @return
	 */
	public static boolean isNameHomotypic(TaxonNameBase name, Taxon taxon) {
		TaxonNameBase taxonName = taxon.getName();
		if (taxonName == null || name == null) {
			return false;
		}
		HomotypicalGroup homotypicGroup = taxonName.getHomotypicalGroup();
		if (homotypicGroup == null) {
			return false;
		}
		if (homotypicGroup.equals(name.getHomotypicalGroup())) {
			return true;
		}
		return false;
	}

	/**
	 * Returns whatever is currently considered the display name for Taxon
	 * objects. Currently titleCache.
	 * 
	 * @param taxon
	 * @return
	 */
	public static String getDisplayName(TaxonBase taxonBase) {
		TaxonNameBase name = taxonBase.getName();
		return getDisplayName(name);
	}

	/**
	 * @param name
	 * @return
	 */
	public static String getDisplayName(TaxonNameBase name) {
		if (name != null) {
			return name.getTitleCache().replace("&", "&&");
		}
		return "";
	}
	
	public static String getDisplayNameWithRef(TaxonBase taxonBase) {
		TaxonNameBase name = taxonBase.getName();
		if (name != null) {
			if (name.getFullTitleCache() == null || name.getFullTitleCache().length() == 0) {
				return CdmUtils.Nz(name.getTitleCache());
			} else {
				return name.getFullTitleCache();
			}
		}
		return "";
	}

	/**
	 * @param oldTaxon
	 * @param newAcceptedTaxon
	 * @param synonymType
	 * @param citation
	 * @param citationMicroReference
	 */
	public static void makeTaxonSynonym(Taxon oldTaxon, Taxon newAcceptedTaxon,
			SynonymRelationshipType synonymType, ReferenceBase citation,
			String citationMicroReference) {
		ITaxonService taxonService = getTaxonService();
		taxonService.makeTaxonSynonym(oldTaxon, newAcceptedTaxon, synonymType,
				citation, citationMicroReference);
	}

	/**
	 * @param searchText
	 * @return
	 */
	public static List<TaxonNameBase> searchNameString(String searchText) {
		List<TaxonNameBase> resultsSet = new ArrayList<TaxonNameBase>();
		resultsSet.addAll(getNameService()
				.getNamesByName(searchText.replace("*", "%")));
		return resultsSet;
	}
	
	public static SearchResult searchTaxaByName(String name) {
		SearchResult searchResult = new SearchResult();
		
		ITaxonServiceConfigurator configurator = new TaxonServiceConfiguratorImpl();
		
		configurator.setMatchMode(MatchMode.ANYWHERE);
		configurator.setSearchString(name.replace("*", "%"));
		configurator.setDoTaxa(true);
		configurator.setDoNamesWithoutTaxa(true);
		configurator.setDoSynonyms(true);
		configurator.setPageNumber(1);
		// TODO currently limit results to 1000 for now
		configurator.setPageSize(1000);
		
		List<IdentifiableEntity> result = getTaxonService().findTaxaAndNames(configurator).getRecords();
		
		searchResult.addAll(result);
		return searchResult;
	}
	
	private static CdmApplicationController getApplicationController() {
		return TaxEditorPlugin.getDefault().getApplicationController();
	}
	
	public static INameService getNameService() {
		return getApplicationController().getNameService();
	}
	
	public static ITaxonService getTaxonService() {
		return getApplicationController().getTaxonService();
	}

	public static IReferenceService getReferenceService() {
		return getApplicationController().getReferenceService();
	}
	
	/**
	 * @param searchText
	 * @param taxon
	 * @return
	 */
	public static Set<TaxonNameBase> getNameByNameForTaxonContext(String searchText, Taxon taxon) {
		Set<TaxonNameBase> resultsSet = new HashSet<TaxonNameBase>();
		resultsSet.addAll(getNameService().getNamesByName
				(searchText.replace("*", "%"), taxon));
		return resultsSet;
	}
		
	/**
	 * Converts a <code>String</code> whose format is either a valid year 
	 * or two valid years with the format "XXXX-XXXX" into a TimePeriod. 
	 *
	 * @see #getValidYear(String yearStr)
	 * @param refYear
	 * @return
	 * @throws NumberFormatException
	 */
	public static TimePeriod convertTimePeriod(String refYear) throws NumberFormatException {
		
		if (refYear == null || ("").equals(refYear)){
			return null;
		}
		
		TimePeriod datePublished = TimePeriod.NewInstance();
		
		// In case format is "xxxx-xxxx"
		String[] years = refYear.split("-");

		// Unlikely case of "xxxx-xxxx-xxxx..."
		if (years.length > 2) {
			throw new NumberFormatException();
		}
		
		// Set startYear
		datePublished.setStartYear(getValidYear(years[0]));

		// Format is "xxxx-xxxx"
		if (years.length == 2) {
			datePublished.setEndYear(getValidYear(years[1]));
		}
		
		return datePublished;
	}
	
	
	/**
	 * Checks whether a <code>String</code> is a valid year between
	 * 1750 and 2030. Throws a <code>NumberFormatException</code> if not.
	 * 
	 * @param yearStr
	 * @return
	 * @throws NumberFormatException
	 */
	public static Integer getValidYear(String yearStr) throws NumberFormatException {
		
		Integer yearInt = null;
		
		// Try casting string - don't catch number format exception
		try {
			yearInt = new Integer(yearStr);
		} catch (ClassCastException e) {
			throw new NumberFormatException();
		}
		
		// Is year in valid range?
		if (yearInt < 1750 || yearInt > 2030) {
			throw new NumberFormatException();
		}
		
		return yearInt;
	}

	/**
	 * Searches for references by string. "%" is used as a wildcard.
	 * 
	 * @param text
	 * @return 
	 */
	public static List getReferencesByTitle(String reference) {
		
		reference = reference.replace("*", "%");
		List resultsList = null;
		try {
			resultsList = getReferenceService().getReferencesByTitle(reference);
		} catch (RuntimeException e) {
			MessageDialog.openError(GlobalController.getShell(), "Search reference error", 
					"Reference search returned an error. This could be a Hibernate concurrency problem. " +
					"Please try saving your work, then searching again.");
			e.printStackTrace();
		}
		return resultsList;
	}
	
	/**
	 * Returns true if <code>taxon</code> belongs to the children, grandchildren,
	 * etc. of <code>checkTaxon</code>.
	 * 
	 * @param taxon
	 * @param checkTaxon
	 * @return
	 */
	public static boolean isTaxonChildOfTaxon(Taxon taxon, Taxon checkTaxon) {
		
		// Traverse all checkTaxon's children
		for (Taxon childTaxon : checkTaxon.getTaxonomicChildren()) {
			
			if (childTaxon.equals(taxon)) {
				return true;
			} else {
				
				// Compare taxon with childTaxon's children
				if (isTaxonChildOfTaxon(taxon, childTaxon)) {
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * @param name
	 * @return
	 */
	public static boolean isNameSupraSpecific(TaxonNameBase name) {
		
		if (name == null || name.getRank() == null) {
			return false;
		}
		
		if (name.getRank().isHigher(Rank.SPECIES())) {
//		if (name.isInfraGeneric() || name.isSupraGeneric() || name.isGenus()) {
			return true;
		} else {
			return false;
		}
	}
}
