// $Id$
/**
* Copyright (C) 2007 EDIT
* European Distributed Institute of Taxonomy 
* http://www.e-taxonomy.eu
* 
* The contents of this file are subject to the Mozilla Public License Version 1.1
* See LICENSE.TXT at the top of this package for the full license terms.
*/

package eu.etaxonomy.taxeditor.annotatedlineeditor;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.apache.log4j.Logger;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.Position;
import org.eclipse.jface.text.source.Annotation;
import org.eclipse.jface.text.source.AnnotationModel;
import org.eclipse.jface.text.source.AnnotationModelEvent;

import eu.etaxonomy.cdm.model.name.BotanicalName;

/**
 * An <code>AnnotationModel</code> which holds <code>LineAnnotation</code>'s.
 * <p>
 * Major difference with <code>AnnotationModel</code> is that <code>removeAnnotation(Annotation annotation,
 * boolean fireModelChanged)</code> adds annotations marked for removal to a collection for further
 * processing, i.e. when the document is saved. This collection is accessed via <code>getDeletedAnnotations()</code>
 * and <code>clearDeletedAnnotations()</code>.
 * 
 * @author p.ciardelli
 * @created 25.06.2009
 * @version 1.0
 */
public class LineAnnotationModel extends AnnotationModel {
	private static final Logger logger = Logger
			.getLogger(LineAnnotationModel.class);
	
	private Comparator<Annotation> comparator;
	private IEntityCreator<?> entityCreator;
	private Set<LineAnnotation> deletedAnnotations = new HashSet<LineAnnotation>();

	private ILineDisplayStrategy lineDisplayStrategy;
	
	public LineAnnotationModel(ILineDisplayStrategy lineDisplayStrategy) {
		super();
		this.lineDisplayStrategy = lineDisplayStrategy;
	}
	
	/**
	 * Changes the annotation's type and fires model changed 
	 * notification. Resulting redraw will cause any change
	 * in annotation's decoration to be displayed. 
	 * 
	 * @param annotation
	 * @param type
	 */
	public void changeAnnotationType(Annotation annotation, String type) {
		annotation.setType(type);
		fireModelChanged(new AnnotationModelEvent(this));
	}
	
	/* (non-Javadoc)
	 * @see org.eclipse.jface.text.source.AnnotationModel#removeAnnotation(org.eclipse.jface.text.source.Annotation, boolean)
	 */
	protected void removeAnnotation(Annotation annotation,
			boolean fireModelChanged) {
		// Set data model element for deletion
		annotation.markDeleted(true);
		deletedAnnotations.add((LineAnnotation) annotation);
		super.removeAnnotation(annotation, fireModelChanged);
	}
		
	/**
	 * @param type
	 */
	public void removeTypeFromAllAnnotations(String type) {
		Iterator<?> iterator = getAnnotationIterator();
		while (iterator.hasNext()) {
			Annotation annotation = ((Annotation) iterator.next());
			if (annotation.getType().equals(type)) {
				annotation.setType(Annotation.TYPE_UNKNOWN);
			}
		}
	}
	
	/**
	 * @param type
	 * @return
	 */
	public Set<LineAnnotation> getAllAnnotationsOfType(String type) {
		Set<LineAnnotation> candidates = new HashSet<LineAnnotation>();
		Iterator<?> iterator = getAnnotationIterator();
		while (iterator.hasNext()) {
			LineAnnotation annotation = ((LineAnnotation) iterator.next());
			if (annotation.getType().equals(type)) {
				candidates.add(annotation);
			}
		}	
		return candidates;	
	}	
	
	/**
	 * @param type
	 * @return
	 */
	public Annotation getFirstAnnotationOfType(String type) {
		Iterator<?> iterator = getAnnotationIterator();
		while (iterator.hasNext()) {
			Annotation annotation = ((Annotation) iterator.next());
			if (annotation.getType().equals(type)) {
				return annotation;
			}
		}		
		return null;
	}
	
	/**
	 * @param line
	 * @param document
	 * @param model
	 * @return
	 */
	public Annotation getAnnotationAtLine(int line, IDocument document) {
		Annotation annotation = null;
		try {
			int offset = document.getLineOffset(line);
			int length = document.getLineLength(line);
			
			// test explicitly add annotation, see if the iterator still returns nothing
//			LineAnnotation newAnnotation = new LineAnnotation(BotanicalName.NewInstance(null), lineDisplayStrategy);
//			Position position = new Position(offset, length);
//			addAnnotation(newAnnotation, position);
			// newly added annotation gets returned
			
			// test whether an annotation iterator w no position returns anything
//			Iterator<?> iterator = getAnnotationIterator();
			// also return nothing
			
			// so annotations which have been added to this annotation model are disappearing
			// model stays the same
			
			Iterator<?> iterator = getAnnotationIterator(offset, length, true, true); // returning nothing
			if (iterator.hasNext()) {
				annotation = (Annotation) iterator.next();
			}
		} catch (BadLocationException e) {
			// do nothing
		}
		return annotation;
	}
		
	/**
	 * @return
	 */
	public List<Annotation> getOrderedAnnotations() {
		List<Annotation> list = new ArrayList<Annotation>();
		Iterator<?> iterator = getAnnotationIterator();
		while (iterator.hasNext()) {
			list.add((Annotation) iterator.next());
		}
		Collections.sort(list, getAnnotationComparator());
		return list;
	}
		
	public void printAnnotations() {
		logger.debug("------------------------");
		logger.debug("Active annotations");
		logger.debug("------------------------");
		List<Annotation> list = getOrderedAnnotations();
		
		for (Annotation annotation : list) {
			logger.debug(
					(annotation.isMarkedDeleted() ?  "DELETED " : "") +
					(((LineAnnotation) annotation).isMarkedAsMerged() ?  "MERGED " : "") +
					(((LineAnnotation) annotation).isDirty() ?  "DIRTY " : "") +
					(((LineAnnotation) annotation).isMarkedAsNew() ?  "NEW " : "") +
					annotation.getText() + ": o " +
					getPosition(annotation).getOffset() + ", l " + 
					getPosition(annotation).getLength());
		}
		logger.debug("------------------------");
		logger.debug("Deleted annotations");
		logger.debug("------------------------");
		for (LineAnnotation annotation : deletedAnnotations) {
			logger.debug(
					(annotation.isMarkedDeleted() ?  "DELETED " : "") +
					(((LineAnnotation) annotation).isMarkedAsMerged() ?  "MERGED " : "") +
					(((LineAnnotation) annotation).isDirty() ?  "DIRTY " : "") +
					(((LineAnnotation) annotation).isMarkedAsNew() ?  "NEW " : "") +
					annotation.getText() + ": o ");			
		}
	}
	
	/**
	 * @param offset
	 * @param length
	 * @return
	 */
	public List<LineAnnotation> getUndeletedAnnotations(int offset, int length) {
		
		List<LineAnnotation> list = new ArrayList<LineAnnotation>();
		
		Iterator<?> iterator = getAnnotationIterator(offset, length, true, true);
		while (iterator.hasNext()) {
			Object next = iterator.next();
			if (next instanceof LineAnnotation && !((Annotation) next).isMarkedDeleted()) {
				list.add((LineAnnotation) next); 
			}
		}
		Collections.sort(list, getAnnotationComparator());
		return list;
	}
	
	public Set<LineAnnotation> getDeletedAnnotations() {
		return deletedAnnotations;
	}
	
	public void clearDeletedAnnotations() {
		deletedAnnotations.clear();
	}
	
	private Comparator<Annotation> getAnnotationComparator() {
		if (comparator == null) {
			this.comparator = new AnnotationComparator();
		}
		return comparator;
	}
	
	class AnnotationComparator implements Comparator<Annotation> {

		/* (non-Javadoc)
		 * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
		 */
		public int compare(Annotation annotation1, Annotation annotation2) {
			if (annotation1.isMarkedDeleted() && !annotation2.isMarkedDeleted()) {
				return 1;
			}
			if (!annotation1.isMarkedDeleted() && annotation2.isMarkedDeleted()) {
				return -1;
			}
			int offset1 = LineAnnotationModel.this.
								getPosition(annotation1).getOffset();
			int offset2 = LineAnnotationModel.this.
								getPosition(annotation2).getOffset();
			if (offset1 > offset2) {
				return 1;
			} else if (offset1 < offset2) {
				return -1;
			} else {
				return 0;
			}
		}	
	}
	
	/**
	 * The annotation model is assigned responsibility for creating new containers
	 * from blocks of text.
	 * 
	 * @param entityCreator
	 */
	public void setEntityCreator(IEntityCreator<?> entityCreator) {
		this.entityCreator = entityCreator;
	}
	
	/**
	 * Creates an annotation without adding it to the model.
	 * 
	 * This object must have its <code>cdmEntityCreator</code> set. 
	 * 
	 * @param text
	 * @return
	 */
	public Annotation createAnnotation(String text) {
		LineAnnotation annotation = null;
		if (entityCreator != null) {
			annotation = new LineAnnotation(entityCreator.createEntity(text), lineDisplayStrategy);
			annotation.markAsNew(true);
		}
		return annotation;
	}
	
	/* (non-Javadoc)
	 * @see org.eclipse.jface.text.source.AnnotationModel#addAnnotation(org.eclipse.jface.text.source.Annotation, org.eclipse.jface.text.Position)
	 */
	@Override
	public void addAnnotation(Annotation annotation, Position position) {
		super.addAnnotation(annotation, position);
	}
	
	public Annotation getAnnotation(Object entity) {
		if (entity != null) {
			Iterator iterator = getAnnotationIterator();
			while (iterator.hasNext()) {
				LineAnnotation annotation = (LineAnnotation) iterator.next();
				if (entity.equals(annotation.getEntity())) {
					return annotation;
				}
			}
		}
		return null;
	}
}