/**
 * 
 */
package eu.etaxonomy.taxeditor.editor.name.handler;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.UUID;

import org.apache.log4j.Logger;
import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.commands.IHandler;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.handlers.HandlerUtil;

import eu.etaxonomy.cdm.api.conversation.ConversationHolder;
import eu.etaxonomy.cdm.model.description.TaxonDescription;
import eu.etaxonomy.cdm.model.name.TaxonNameBase;
import eu.etaxonomy.cdm.model.reference.ReferenceBase;
import eu.etaxonomy.cdm.model.taxon.SynonymRelationship;
import eu.etaxonomy.cdm.model.taxon.SynonymRelationshipType;
import eu.etaxonomy.cdm.model.taxon.Taxon;
import eu.etaxonomy.cdm.model.taxon.TaxonNode;
import eu.etaxonomy.cdm.model.taxon.TaxonRelationship;
import eu.etaxonomy.cdm.model.taxon.TaxonRelationshipType;
import eu.etaxonomy.taxeditor.dialogs.filteredSelection.FilteredTaxonNodeSelectionDialog;
import eu.etaxonomy.taxeditor.editor.EditorUtil;
import eu.etaxonomy.taxeditor.propertysheet.name.TaxonBasePropertySource;
import eu.etaxonomy.taxeditor.store.CdmStore;

/**
 * @author p.ciardelli
 *
 */
public class ChangeAcceptedTaxonToSynonymHandler extends AbstractHandler
		implements IHandler {
	private static final Logger logger = Logger
			.getLogger(ChangeAcceptedTaxonToSynonymHandler.class);

	/* (non-Javadoc)
	 * @see org.eclipse.core.commands.IHandler#execute(org.eclipse.core.commands.ExecutionEvent)
	 */
	public Object execute(ExecutionEvent event) throws ExecutionException {
		ConversationHolder conversation = CdmStore.NewTransactionalConversation();
		conversation.bind();
		
		ISelection menuSelection = HandlerUtil.getActiveMenuSelection(event);
		TaxonBasePropertySource taxonBasePropertySource = 
				(TaxonBasePropertySource) ((StructuredSelection) menuSelection).getFirstElement();
		Taxon oldAcceptedTaxon = (Taxon) taxonBasePropertySource.getTaxonBase();
		
		// Choose new accepted taxon	
		List<UUID> excludeTaxa = new ArrayList<UUID>();
		excludeTaxa.add(oldAcceptedTaxon.getUuid());
		Taxon newAcceptedTaxon = FilteredTaxonNodeSelectionDialog.selectTaxonNode(HandlerUtil.getActiveShell(event), "Choose the accepted taxon", excludeTaxa, null).getTaxon();
		
		
		if (newAcceptedTaxon == null) {
			return null;
		}
		
		try {
			
			Taxon newTaxon = (Taxon) CdmStore.getTaxonService().getTaxonByUuid(newAcceptedTaxon.getUuid());
			Taxon oldTaxon = (Taxon) CdmStore.getTaxonService().getTaxonByUuid(oldAcceptedTaxon.getUuid());
			
			// Check for multiple nodes
			if (newTaxon.getTaxonNodes().size() > 1 || oldTaxon.getTaxonNodes().size() > 1) {
				MessageDialog.openInformation(HandlerUtil.getActiveShell(event), 
					"Taxon implemented in multiple trees", 
					"One or both of these taxa is implemented in multiple taxonomic trees.");
				return null;
			}
			
			// Close editor
			IEditorPart editor = HandlerUtil.getActiveEditor(event);
			boolean proceed = HandlerUtil.getActiveWorkbenchWindowChecked(event).getActivePage().
								closeEditor(editor, true);
			if (!proceed) {
				return null;
			}
			
			makeTaxonSynonym(oldTaxon, newTaxon, null, null, null);
				
			conversation.commit();
			Set<TaxonNode> nodes = newTaxon.getTaxonNodes();
			UUID uuid = nodes.iterator().next().getUuid();
			
			EditorUtil.open(uuid);
						
		} catch (PartInitException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		return null;
	}

	/**
	 * @param oldTaxon
	 * @param newTaxon
	 */
	private void makeTaxonSynonym(Taxon oldTaxon, Taxon newAcceptedTaxon, SynonymRelationshipType synonymType, ReferenceBase citation, String citationMicroReference) {
		if (oldTaxon == null || newAcceptedTaxon == null || oldTaxon.getName() == null){
			return;
		}
		
		// Move oldTaxon to newTaxon
		TaxonNameBase<?,?> synonymName = oldTaxon.getName();
		if (synonymType == null){
			if (synonymName.isHomotypic(newAcceptedTaxon.getName())){
				synonymType = SynonymRelationshipType.HOMOTYPIC_SYNONYM_OF();
			}else{
				//TODO synonymType 
				synonymType = SynonymRelationshipType.HETEROTYPIC_SYNONYM_OF();
			}
		}
		SynonymRelationship synRel = newAcceptedTaxon.addSynonymName(synonymName, synonymType, citation, citationMicroReference);
		
		//Move Synonym Relations to new Taxon
		for(SynonymRelationship synRelation : oldTaxon.getSynonymRelations()){
			synRelation.setAcceptedTaxon(newAcceptedTaxon);
		}

		//Move Taxon RelationShips to new Taxon
		Set<TaxonRelationship> removableTaxonRels = new HashSet<TaxonRelationship>();
		for(TaxonRelationship taxonRelation : oldTaxon.getTaxonRelations()){
			//CHILDREN
			if (taxonRelation.getType().equals(TaxonRelationshipType.TAXONOMICALLY_INCLUDED_IN())){
				if (taxonRelation.getFromTaxon() == oldTaxon){
					taxonRelation.setFromTaxon(newAcceptedTaxon);
				}else if(taxonRelation.getToTaxon() == oldTaxon){
					taxonRelation.setToTaxon(newAcceptedTaxon);
				}else{
					logger.warn("Taxon is not part of its own Taxonrelationship");
				}
			}
			//MISAPPLIED NAMES
			if (taxonRelation.getType().equals(TaxonRelationshipType.MISAPPLIED_NAME_FOR())){
				if (taxonRelation.getFromTaxon() == oldTaxon){
					taxonRelation.setFromTaxon(newAcceptedTaxon);
				}else if(taxonRelation.getToTaxon() == oldTaxon){
					taxonRelation.setToTaxon(newAcceptedTaxon);
				}else{
					logger.warn("Taxon is not part of its own Taxonrelationship");
				}
			}
			//Concept Relationships
			//FIXME implement
			logger.warn("Transfer of concept relationships not implemented yet");
		}
		
		
		//Move Descriptions to new Taxon
		Set<TaxonDescription> descriptions = new HashSet<TaxonDescription>();
		for(TaxonDescription taxDescription : oldTaxon.getDescriptions()){
			descriptions.add(taxDescription);
		}						
		for(TaxonDescription taxDescription : descriptions){
			newAcceptedTaxon.addDescription(taxDescription);
		}


		oldTaxon.getTaxonNodes().iterator().next().delete();
		
	}
	
}