// $Id$
/**
* Copyright (C) 2007 EDIT
* European Distributed Institute of Taxonomy 
* http://www.e-taxonomy.eu
* 
* The contents of this file are subject to the Mozilla Public License Version 1.1
* See LICENSE.TXT at the top of this package for the full license terms.
*/

package eu.etaxonomy.taxeditor.store.operations;

import org.eclipse.core.commands.ExecutionException;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

import eu.etaxonomy.cdm.model.name.HomotypicalGroup;
import eu.etaxonomy.cdm.model.name.NonViralName;
import eu.etaxonomy.cdm.model.taxon.SynonymRelationship;
import eu.etaxonomy.cdm.model.taxon.SynonymRelationshipType;
import eu.etaxonomy.cdm.model.taxon.Taxon;
import eu.etaxonomy.taxeditor.operations.CreateSynonymInExisitingHomotypicalGroupOperation;

/**
 * @author n.hoffmann
 * @created 07.04.2009
 * @version 1.0
 */
public class CreateSynonymInExisitingHomotypicalGroupOperationTest extends AbstractTaxeditorOperationTest {

	private String newSynonymText;

	private HomotypicalGroup homotypicalGroup;


	private NonViralName<?> taxonName;
	
	/**
	 * @throws java.lang.Exception
	 */
	@Before
	public void setUp() throws Exception {
		
		taxonName = NonViralName.NewInstance(null);
		taxon = Taxon.NewInstance(taxonName, null);
		
		homotypicalGroup = HomotypicalGroup.NewInstance();
		newSynonymText = "";
		
		operation = new CreateSynonymInExisitingHomotypicalGroupOperation("Create Synonym In Existing Homotypical Group",
							undoContext, taxon, homotypicalGroup, newSynonymText, postOperation);
	}

	/**
	 * Test method for {@link eu.etaxonomy.taxeditor.operations.CreateSynonymInExisitingHomotypicalGroupOperation#execute(org.eclipse.core.runtime.IProgressMonitor, org.eclipse.core.runtime.IAdaptable)}.
	 * 
	 * Homotypic group is the Taxons homotypic group, so we expect the synonym to be homotypic to the accepted taxon.
	 * 
	 * @throws ExecutionException 
	 */
	@Test
	public void testExecuteTaxonHomotypicGroup() throws ExecutionException {		
		homotypicalGroup.addTypifiedName(taxonName);
		operation.execute(monitor, info);
		
		Assert.assertTrue("Taxon should have a synonym now.", taxon.getSynonyms().size() > 0);
		Assert.assertTrue("Taxon should have a homotypic synonym now", taxon.getHomotypicSynonymsByHomotypicGroup().size() > 0);
	}
	
	/**
	 * Test method for {@link eu.etaxonomy.taxeditor.operations.CreateSynonymInExisitingHomotypicalGroupOperation#execute(org.eclipse.core.runtime.IProgressMonitor, org.eclipse.core.runtime.IAdaptable)}.
	 * 
	 * Homotypic group is not the accepted taxons. Heterotypic synonym should be created.
	 * 
	 * @throws ExecutionException 
	 */
	@Test
	public void testExecuteSynonymHomotypicGroup() throws ExecutionException {		
		// test heterotypic synonym
		operation.execute(monitor, info);
		
		Assert.assertTrue("Taxon should have a synonym now.", taxon.getSynonyms().size() > 0);
		Assert.assertTrue("Taxon should have a homotypic group", taxon.getHomotypicSynonymyGroups().size() > 0);
		Assert.assertEquals("Synonym relationship should be heterotypic", SynonymRelationshipType.HETEROTYPIC_SYNONYM_OF(), taxon.getSynonymRelations().toArray(new SynonymRelationship[0])[0].getType());
	}
	

	/**
	 * Test method for {@link eu.etaxonomy.taxeditor.operations.CreateSynonymInExisitingHomotypicalGroupOperation#undo(org.eclipse.core.runtime.IProgressMonitor, org.eclipse.core.runtime.IAdaptable)}.
	 * @throws ExecutionException 
	 */
	@Test
	public void testUndoTaxonHomotypicGroup() throws ExecutionException {
		homotypicalGroup.addTypifiedName(taxonName);
		
		operation.execute(monitor, info);
		operation.undo(monitor, info);
		
		Assert.assertTrue("There should not be synonym relationships for the taxon", taxon.getSynonyms().size() == 0);
	}
	
	/**
	 * Test method for {@link eu.etaxonomy.taxeditor.operations.CreateSynonymInExisitingHomotypicalGroupOperation#undo(org.eclipse.core.runtime.IProgressMonitor, org.eclipse.core.runtime.IAdaptable)}.
	 * @throws ExecutionException 
	 */
	@Test
	public void testUndoSynonymHomotypicGroup() throws ExecutionException {
		operation.execute(monitor, info);
		operation.undo(monitor, info);
		
		Assert.assertTrue("There should not be synonym relationships for the taxon", taxon.getSynonyms().size() == 0);
	}

	/**
	 * Test method for {@link eu.etaxonomy.taxeditor.operations.CreateSynonymInExisitingHomotypicalGroupOperation#redo(org.eclipse.core.runtime.IProgressMonitor, org.eclipse.core.runtime.IAdaptable)}.
	 * @throws ExecutionException 
	 */
	@Test
	public void testRedoTaxonHomotypicGroup() throws ExecutionException {
		homotypicalGroup.addTypifiedName(taxonName);
		
		operation.execute(monitor, info);
		operation.undo(monitor, info);		
		operation.redo(monitor, info);
		
		Assert.assertTrue("Taxon should have a synonym now.", taxon.getSynonyms().size() > 0);
		Assert.assertTrue("Taxon should have a homotypic synonym now", taxon.getHomotypicSynonymsByHomotypicGroup().size() > 0);
	}
	
	/**
	 * Test method for {@link eu.etaxonomy.taxeditor.operations.CreateSynonymInExisitingHomotypicalGroupOperation#redo(org.eclipse.core.runtime.IProgressMonitor, org.eclipse.core.runtime.IAdaptable)}.
	 * @throws ExecutionException 
	 */
	@Test
	public void testRedoSynonymHomotypicGroup() throws ExecutionException {
		operation.execute(monitor, info);
		operation.undo(monitor, info);		
		operation.redo(monitor, info);
		
		Assert.assertTrue("Taxon should have a synonym now.", taxon.getSynonyms().size() > 0);
		Assert.assertTrue("Taxon should have a homotypic group", taxon.getHomotypicSynonymyGroups().size() > 0);
		Assert.assertEquals("Synonym relationship should be heterotypic", SynonymRelationshipType.HETEROTYPIC_SYNONYM_OF(), taxon.getSynonymRelations().toArray(new SynonymRelationship[0])[0].getType());

	}
}
